#!/usr/bin/perl
#  categories.pl - functions to implement categories
#
#  Written by Curtis Olson.  Started September 29, 1994.
#
#  Copyright (C) 1994 - 1999  Curtis L. Olson  - curt@me.umn.edu
#  Portions Copyright 2000 Eugene Magnier - eugene@cfht.hawaii.edu
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# groups.pl,v 0.11 2001/08/12 20:47:36 briank Exp

package CBB;

use strict;    # don't take no guff

# Global variables

# %CBB::ACCTS - an associative array of accounts
# @CBB::ACCOUNTS - a sorted list of account keys (for traversing the cat list)
# $CBB::sorted_accts - specifies whether the list in @CBB::CATKEYS is valid
# %CBB::BALS - an associative array of account info

# initialize the accounts list
sub init_accts {
    # out: result

    %CBB::ACCTS = ();
    @CBB::ACCOUNTS = ();
    $CBB::sorted_accts = 0;

    return "ok";
}


# set @CBB::CATKEYS = sorted list of transaction keys
sub sort_accts {
    $CBB::sorted_accts = 1;

    print DEBUG "sort_accts()\n" if $CBB::debug;
    @CBB::ACCOUNTS = sort(keys %CBB::ACCTS);
}


# add/change an account in the list
# filename should be just file, or relative to group
sub edit_acct {
    # in: account
    # out: account

    my($line) = @_;
    my($key, $file, $desc, $curbal, $endbal, $clearbal) = split(/\t/, $line);

    $CBB::sorted_accts = 0;

    # adds new entry if non-existent
    $curbal = 0.0 if( !$curbal );
    $endbal = 0.0 if( !$endbal );
    $clearbal = 0.0 if( !$clearbal );
    $CBB::ACCTS{$key} = "$file\t$desc\t$curbal\t$endbal\t$clearbal";

    print DEBUG "edit_acct: $line\n" if $CBB::debug;

    return "$line";
}

# rename/move an account in the list
# Does not save new changes
# CAUTION: May load new group!  Make sure current transactions have been saved!
sub rename_acct {
    # in: oldname, newname, newgroup, desc  
    # out: account

    my($line) = @_;
    my($okey, $key, $newgroup, $desc) = split(/\t/, $line);
    my $file = $key;

    $CBB::sorted_accts = 0;

    # error check - does key exist
    $line = find_acct ($okey);
    if( $line eq "none" ) { return "error"; }
    my($ofile, $odesc, $ocurbal, $oendbal, $oclearbal) = split(/\t/, $line);
    $ocurbal = 0.0 if( !$ocurbal );
    $oendbal = 0.0 if( !$oendbal );
    $oclearbal = 0.0 if( !$oclearbal );

    # delete old account key from lists of accounts, account keys
    delete $CBB::ACCTS{$okey};
    for( my $i=0; $i<=$#CBB::ACCOUNTS; ++$i ) {
      if( "$CBB::ACCOUNTS[$i]" eq "$okey" ) {
        splice( @CBB::ACCOUNTS, $i, 1 );
        last;
      }
    }
    
    # Remove entry from category file
    &delete_cat_account( "$CBB::group\t$okey" );

    # If group change, load new group info first!
    if( "$newgroup" ne "$CBB::group" ) {
      &save_accts;
      &save_cats( $CBB::path );
      &load_accts( "$CBB::path\t$newgroup" );
    }
    # Group info
    $CBB::ACCTS{$key} = "$file\t$desc\t$ocurbal\t$oendbal\t$oclearbal";
    push( @CBB::ACCOUNTS, $key );
    $CBB::current_file = $file;

    # Category info
    &insert_cat_account( "$CBB::group\t$key\t$desc" );

    print DEBUG "rename_acct: $line\n" if $CBB::debug;

    return "ok";
}

# add/change an account in the list
sub add_acct {
    # in: account
    # out: account

    my($line) = @_;
    $line = edit_acct ($line);

    return "$line";
}

# delete an account from the list
sub delete_acct {
    # in: category

    my($line) = @_;
    my($key, $file, $desc, $bal, $cbal) = split(/\t/, $line);

    $CBB::sorted_accts = 0;

    delete $CBB::ACCTS{$key};

    print DEBUG "delete_acct:  $line\n" if $CBB::debug;

    return "$line";
}

# returns the entire account list.
sub all_accts {
    # out: category list

    my($key);

    $| = 0;				# turn off buffer flushing

    if ($CBB::sorted_accts == 0) { &sort_accts(); }

    foreach $key (@CBB::ACCOUNTS) {
	print ("$key\t$CBB::ACCTS{$key}\n");
    }

    $| = 1;				# turn buffer flushing back on

    return "none";
}

sub current_acct {

    print DEBUG "current acct: $CBB::current_file" if $CBB::debug;
    my($line) = find_acct_file ($CBB::current_file);
    print DEBUG "line: $line" if $CBB::debug;

    return ($line);
}
    

# find filename of named account
sub find_acct {
    # in: acct name
    # out: acct filename

    my($key) = @_;
    my($line, $name);

    print DEBUG "looking for key: $key...\n" if $CBB::debug;

    # strip whitespace from front & back of $key:
    ($key) = $key =~ /^\s*(.*)$/;
    while ( $key =~ /\s$/ ) {
	chop ($key);
    }
    
    $line = $CBB::ACCTS{$key};
    if ($line eq "") { return "none"; }

    return ($line);
}

# find account given filename
sub find_acct_file {
    # in: acct file
    # out: acct data

    my($input) = @_;
    my($line, $file, $key, $junk);

    if ($CBB::sorted_accts == 0) { &sort_accts(); }

    foreach $key (@CBB::ACCOUNTS) {
	$line = $CBB::ACCTS{$key};
	($file, $junk, $junk, $junk) = split (/\t/, $line);
	if ($file eq $input) {
	    return ("$key\t$line");
	}
    }
    return ("error");

}

# load an account list from a group file, given root path and group name 
# minimal error checking:
#  1) does the file exist?
#  2) does the first line match the standard form?
# it would be nice the validate each line - are all fields available?
sub load_accts {
    # in: root path, group name 
    # out: result

    my($rootPath, $file) = split( /\t/, $_[0] );
    my($line);

    $CBB::sorted_catkeys = 0;


    $CBB::group = $file;
    $CBB::path = $rootPath;

    # Return error if group file doesn't already exist
    my $filename = "$CBB::path/$CBB::group/$CBB::group.grp";
    print DEBUG "load_accts(): looking for $filename\n" if $CBB::debug;
    if( ! -e $filename ) {
      print DEBUG "load_accts(): $filename doesn't exist\n" if $CBB::debug;
      return "error 1"; 
    }

    if( !open(FILE, "<$filename") ) {
      print DEBUG "load_accts(): can't open $filename\n" if $CBB::debug;
      return "error 1";
    }
    print DEBUG "load_accts(): opened $filename\n" if $CBB::debug;

    $line = <FILE>;
    # Must be a group file!
    if (! ($line =~ /# CBB-Group File/)) { return "error 2"; }
    print DEBUG "load_accts(): $filename is a group file!\n" if $CBB::debug;

    while ($line = <FILE> ) {
	chop ($line);
	if ( $line =~ m/^#/ ) {
	     if ( $line =~ m/^#\s*CATEGORIES\s*=/ ) {
		  ($CBB::categories) = $line =~ /^#\s*CATEGORIES\s*=\s*(.*)$/;
		  # print STDERR "cat: $CBB::categories\n";
	      }
	 } else {
	     add_acct ($line);
	 }

    }
    close (FILE);
    return "ok";

}

# save account list to group file
sub save_accts {
    # out: result

    my($key);

    print DEBUG "save_accts(): group=$CBB::group\n";
    if ($CBB::group eq "") { return "error"; }

    if ($CBB::sorted_accts == 0) { &sort_accts(); }

    # Does group already exist?  If not, create group subdir
    my $pathname = "$CBB::path/$CBB::group";
    if( ! -e $pathname ) {
      mkdir $pathname;
    }

    my $filename = "$pathname/$CBB::group.grp";
    print DEBUG "save_accts(): Opening $filename\n";
    open (FILE, ">$filename") || return "error";

    print FILE "# CBB-Group File\n";
    print FILE "#\n";
    print FILE "# CATEGORIES = $CBB::group\n";
    print FILE "# values on each line are:\n";
    print FILE "# (key) (file) (description) (current balance) (ending balance) (cleared balance)\n";
    print FILE "#\n";
    
    foreach $key (@CBB::ACCOUNTS) {
	print FILE "$key\t$CBB::ACCTS{$key}\n";
    }

    close (FILE);

    # look in accounts for method to make save atomic.


    return "ok";
}

# make a new group
sub make_group {
    # in: group-name
    # out: result
    
    my($rootPath, $name) = split( /\t+/, $_[0] );

    print DEBUG "Make group $name\n" if $CBB::debug;

    init_accts ();

    $CBB::group = $name;
    $CBB::path = $rootPath;

    save_accts ();

    return "ok";
}

&init_accts ();

# there are still some problems here:
# 1) the group name is currently the filename.  this will only 
#    work with full paths, which makes things too rigid.
