#!/usr/bin/perl
#  import.pl - functions to implement importing from other formats
#
#  Written by Curtis Olson.  Started August 25, 1994.
#
#  Copyright (C) 1994 - 1999  Curtis L. Olson  - curt@me.umn.edu
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# import.pl,v 0.10 2001/07/13 04:05:59 briank Exp


package CBB;

use strict;   # don't take no guff


# @INC already contains "." by default; no need to modify @INC here. 

require "common.pl";

# load the data from a dbm file
sub load_dbm_trans {
    # in: file base name
    # out: result

    my($file) = @_;
    print DEBUG "Loading the dbm format file: $file\n" if $CBB::debug;

    if ( -e "$file" ) {
	$CBB::current_file = $file;
	$CBB::sorted_keys = 0;
	$CBB::calced = 0;

	dbmclose(%CBB::TRANS);
	dbmopen(%CBB::TRANS, &file_root($file), 0666) || return "error";
	
	# test to see if this file is <tab> delimited
	&sort_keys();
	# never ever call calc_trans() at this point (or call something that
	# calls it
	if (defined($CBB::TRANS{$CBB::KEYS[0]}) && 
	    !($CBB::TRANS{$CBB::KEYS[0]} =~ m/\t/) ) {
	    print DEBUG "'$CBB::TRANS{$CBB::KEYS[0]}' = old version of CBB dbm file\n"
		if $CBB::debug;
	    return "error - old version of CBB dbm file";
	} else {
	    print DEBUG "valid txn: '$CBB::TRANS{$CBB::KEYS[0]}'\n" if $CBB::debug;
        }

	return "ok";
    } else {
	return "error";
    }
}


# load data from a CBB format file
sub load_cbb {
    # in: file base name
    # out: result

    my($file) = @_;
    $CBB::sorted_keys = 0;
    $CBB::calced = 0;

    open(LOAD, "<$file") || return "error";

    while ( <LOAD> ) {
	if ( m/^\s*#/ ) {
	    # toss the comment (any line whose 1st non-whitespace character is
	    #                   the pound sign.)
	} else {
	    chop;
	    &create_trans($_);
	}
    }

    close(LOAD);

    return "ok";
}

# import a quicken interchange file for a single account (.qif)
sub import_qif_account {
    # in: file
    # out: result

    my($file) = @_;
    my($date, $check, $desc, $debit, $credit, $cat, $split, $com, $cleared);
    my($amt, $found_split_com, $split_amt);
    my($day, $month, $year);
    my ($Ntype, $Naccount, $answer, @lines);
    my($iend, $incat, $splitsub, $insplit, $OS);
    local ($/);

    $CBB::sorted_keys = 0;
    $CBB::calced = 0;

    $OS = OSsource ($file);
    if ($OS eq "UNIX") {
	$/ = "\n";
    }
    if ($OS eq "MAC") {
	$/ = "\r";
    }
    if ($OS eq "DOS") {
	# check on this one:
	$/ = "\r\n";
    }
    if ($OS eq "UNKNOWN") {
    }

    open(QIF, "<$file");
    @lines = <QIF>;
    close (QIF);
    foreach ( @lines ) {
	chop;
	if ($OS eq "DOS") { chop; }
    }
    $/ = "\n";

    $Ntype = 0;
    $Naccount = 0;
    # check that this file contains only 1 account:
    foreach (@lines) {
	if ( m/^\!/ ) {
	    if ( m/^\!Account/ ) { 
		$Naccount ++;
		next;
	    }
	    if ( m/^\!Type/ ) { 
		$Ntype ++;
		next;
	    }
	    next;
	} 
    }
    if ($Ntype > 1) {
	print "error more than one account - load as a group\n";
	return "error";
    }

    ($date, $check, $desc, $debit, $credit, $cat, $split, $com, 
     $cleared) = ("", "", "", "", "", "", "", "", "");

    foreach ( @lines ) {
	# s/:/-/g;		# eliminate our delimiter characters from
	s/\|/-/g;		# the import file.
	if ( m/^\!/ ) {
	    # Type
	    next;
	} 
	if ( m/^D/ ) {
	    # Date
	    ($month, $day, $year) = /^D(\d+)\s*\/\s*(\d+)\s*\/\s*(\d+)/;
	    $year = &pad($year);
	    $month = &pad($month);
	    $day = &pad($day);
	    $date = "$year$month$day";
	    # print "$date\n";
	    next;
	} 
	if ( m/^T/ ) {
	    # Transaction Amount
	    s/,//g;		# remove , from numbers (i.e. thousands)
	    $amt = substr($_,1);
	    if ($amt >= 0) {
		$credit = $amt;
		$debit = 0;
	    } else {
		$debit = substr($amt,1); # remove the '-' to make amt >= 0
		$credit = 0;
	    }
	    next;
	} 
	if ( m/^C/ ) {
	    # Cleared
	    $cleared = substr($_,1);
	    $cleared = 'x' if ($cleared eq 'X');
	    next;
	} 
	if ( m/^N/ ) {
	    # check Number
	    $check = substr($_,1);
	    # print "Check # = $check\n";
	    next;
	} 
	if ( m/^P/ ) {
	    # Description
	    $desc = substr($_,1);
	    # print "$desc\n";
	    next;
	} 
	if ( m/^L/ ) {
	    # category
	    $cat = substr($_,1);
	    #print "Category = $cat\n";
	    next;
	    # need to validate the account transfer categories in the file, after imported
	} 
	if ( m/^S/ ) {
	    # split category
	    $splitsub = substr($_,1);
	    if ($split eq "") {
		# first split
		$split = "|".$splitsub."|";
	    } else {
		# not first split :)
		$split = $split.$splitsub."|";
	    }
	    $found_split_com = 0;
	    next;
	} 
	if ( m/^E/ ) {
	    # split comment
	    $split = $split.substr($_,1)."|";
	    $found_split_com = 1;
	    next;
	} 
	if ( m/^\$/ ) {
	    # split amount
	    if ( ! $found_split_com ) {
		$split = $split . "|";
	    }
	    $split_amt = substr($_,1);
	    $split_amt =~ s/\,//g;
	    $split = $split . $split_amt . "|";
	    next;
	} 
	if ( m/^M/ ) {
	    # Comment
	    $com = substr($_,1);
	    next;
	} 
	if ( m/^\^/ ) {
	    #print "End of record\n";
	    if ($split ne "") {
		$cat = $split;
	    }
	    $answer = create_trans("$date\t$check\t$desc\t$debit\t$credit\t$cat\t$com\t$cleared\t0.00" );
	    ($date, $check, $desc, $debit, $credit, $cat, $split, $com, $cleared) = ("", "", "", "", "", "", "", "", "");
	    next;
	} 
	if ( m/^A(\S*)\s*$/ ) {
	    if ($1 ne "") {
		# print "address line: $_\n";
	    }
	    next;
	} 
	if ( m/^U/ ) {
	    # toss U lines
	    # added by Eugene Magnier 7/11/2000
	    next;
	} 
	if ( $_ eq "" ) {
	    # toss empty lines ...
	    next;
	}
	# print "unknown data: $_\n";

    }

    return "ok";
}

# import a quicken interchange file for a group
# need to have a defined group already
sub import_qif_group {
    # in: file
    # out: result

    my($file) = @_;
    my($date, $check, $desc, $debit, $credit, $cat, $split, $com, $cleared);
    my($amt, $found_split_com, $split_amt);
    my($day, $month, $year);
    my ($Ntype, $Naccount, $answer, @lines, $Nmatch);
    my($iend, $incat, $splitsub, $insplit, $OS);
    my($autoswitch, $account, $name, $type, $found, $line, $junk);
    local ($/);

    $CBB::sorted_keys = 0;
    $CBB::calced = 0;

    $OS = OSsource ($file);
    if ($OS eq "UNIX") {
	$/ = "\n";
    }
    if ($OS eq "MAC") {
	$/ = "\r";
    }
    if ($OS eq "DOS") {
	# check on this one:
	$/ = "\r\n";
    }
    if ($OS eq "UNKNOWN") {
    }

    open(QIF, "<$file");
    @lines = <QIF>;
    close (QIF);
    foreach ( @lines ) {
	chop;
	if ($OS eq "DOS") { chop; }
    }
    $/ = "\n";

    # find the account names from AutoSwitch section
    # make the appropriate file
    $autoswitch = 0;
    $account = 0;
    $Naccount = 0;
    foreach (@lines) {
	if ( m/^\!Option:AutoSwitch/ ) { 
	    $autoswitch ++;
	    next;
	}
	if ( m/^\!Clear:AutoSwitch/ ) { 
	    last;
	}
	if ( m/^\!Account/ ) { 
	    $account = 1;
	    next;
	}
	if ( m/^N/ ) { 
	    $name = substr($_,1);
	    next;
	}
	if ( m/^T/ ) { 
	    $type = substr($_,1);
	    next;
	}
	if ( m/^D/ ) { 
	    $desc = substr($_,1);
	    next;
	}
	if ( m/^\^/ ) { 
	    $file = sprintf "file%02d.cbx",$Naccount;
	    $line = make_acct ("$name\t$file\t$desc");
	    $account = 0;
	    $Naccount ++;
	    ($name, $file, $desc) = ("", "", "");
	    next;
	}
	next;
    }

    ($date, $check, $desc, $debit, $credit, $cat, $split, $com, 
     $cleared) = ("", "", "", "", "", "", "", "", "");

    $autoswitch = 0;
    $account = 0;
    $found = 1;
    my($i) = 0;
    foreach ( @lines ) {
	# skip lines until we've gone past the Clear:autoswitch
	if ( m/^\!Option:AutoSwitch/ ) { 
	    $autoswitch ++;
	    next;
	}
	if ( m/^\!Clear:AutoSwitch/ ) { 
	    $autoswitch ++;
	    next;
	}
	if ($autoswitch < 3) { next; }

	if ( m/^\!Account/ ) { 
	    if ($account > 0) {
		save_trans ("$CBB::current_file");
	    }
	    $found = 0;
	    $account ++;
	    $CBB::calced = 0;
	    $CBB::sorted_keys = 0;
	    next;
	}
	if ( m/^N/ ) { 
	    $name = substr($_,1);
	    next;
	}
	if ( m/^\!Type/ ) { 
	    # we now know which account we are dealing with, load it in
	    $line = find_acct ($name);
	    ($file, $junk) = split (/\t/, $line);
	    $line = load_trans ($file);
	    $found = 1;
	    next;
	}
	# until we've loaded in the account details, skip
	# we are skipping the repetative values of D & T
	if (! $found) { next; }
	
	# s/:/-/g;		# eliminate our delimiter characters from
	s/\|/-/g;		# the import file.
	if ( m/^\!/ ) {
	    # Type
	    print "Type $_\n";
	    next;
	} 
	if ( m/^D/ ) {
	    # Date
	    unless (($month, $day, $year) = /^D(\d+)\s*\/\s*(\d+)\s*\/\s*(\d+)/) {
		($month, $day, $year) = /^D(\d*)\/(..).(..)/;
	    }		
	    $year = &pad($year);
	    $month = &pad($month);
	    $day = &pad($day);
	    $date = "$year$month$day";
	    next;
	} 
	if ( m/^T/ ) {
	    # Transaction Amount
	    s/,//g;		# remove , from numbers (i.e. thousands)
	    $amt = substr($_,1);
	    if ($amt >= 0) {
		$credit = $amt;
		$debit = 0;
	    } else {
		$debit = substr($amt,1); # remove the '-' to make amt >= 0
		$credit = 0;
	    }
	    next;
	} 
	if ( m/^C/ ) {
	    # Cleared
	    $cleared = substr($_,1);
	    $cleared = 'x' if ($cleared eq 'X');
	    next;
	} 
	if ( m/^N/ ) {
	    # check Number
	    $check = substr($_,1);
	    # print "Check # = $check\n";
	    next;
	} 
	if ( m/^P/ ) {
	    # Description
	    $desc = substr($_,1);
	    # print "$desc\n";
	    next;
	} 
	if ( m/^L/ ) {
	    # category
	    $cat = substr($_,1);
	    #print "Category = $cat\n";
	    next;
	    # need to validate the account transfer categories in the file, after imported
	} 
	if ( m/^S/ ) {
	    # split category
	    $splitsub = substr($_,1);
	    if ($split eq "") {
		# first split
		$split = "|".$splitsub."|";
	    } else {
		# not first split :)
		$split = $split.$splitsub."|";
	    }
	    $found_split_com = 0;
	    next;
	} 
	if ( m/^E/ ) {
	    # split comment
	    $split = $split.substr($_,1)."|";
	    $found_split_com = 1;
	    next;
	} 
	if ( m/^\$/ ) {
	    # split amount
	    if ( ! $found_split_com ) {
		$split = $split . "|";
	    }
	    $split_amt = substr($_,1);
	    $split_amt =~ s/\,//g;
	    $split = $split . $split_amt . "|";
	    next;
	} 
	if ( m/^M/ ) {
	    # Comment
	    $com = substr($_,1);
	    next;
	} 
	if ( m/^\^/ ) {
	    #print "End of record\n";
	    if ($split ne "") {
		$cat = $split;
	    }
	    $answer = create_trans("$date\t$check\t$desc\t$debit\t$credit\t$cat\t$com\t$cleared\t0.00" );
	    ($date, $check, $desc, $debit, $credit, $cat, $split, $com, $cleared) = ("", "", "", "", "", "", "", "", "");
	    next;
	} 
	if ( m/^A(\S*)\s*$/ ) {
	    if ($1 ne "") {
		# print "address line: $_\n";
	    }
	    next;
	} 
	if ( m/^U/ ) {
	    # toss U lines
	    # added by Eugene Magnier 7/11/2000
	    next;
	} 
	if ( $_ eq "" ) {
	    # toss empty lines ...
	    next;
	}
	# print "unknown data: $_\n";
    }

    if ($account != $Naccount) {
	print "error: didn't find enough account files!\n";
    }

    save_trans ("$CBB::current_file");
    return "ok";
}

# import cbb 0.8 version file
# there must already be an account open.  don't init the 
# current account
sub import_cbb8 {
    # in: filename
    # out: result

    my($file) = @_;
    my($file_version) = "";
    my($junk);

    $CBB::sorted_keys = 0;
    $CBB::calced = 0;

    print DEBUG "Loading the cbb format file: $file\n" if $CBB::debug;

    if ( $CBB::decrypt ne "" ) {
	open(LOAD, "$CBB::decrypt < $file|") || return "error";
    } else {
    	open(LOAD, "<$file") || return "error";
    }

    while ( <LOAD> ) {
	if ( m/^#/ ) {
	    # toss the comment (but first check for any goodies.)
	    if ( m/version/i ) {
		($junk, $junk, $junk, $file_version) = split;
		print DEBUG "Data file version = $file_version\n" if $CBB::debug;
	    }
	} else {
	    if ( $file_version eq "") {
		print DEBUG "no data file version, file encrypted ?" if $CBB::debug;
		close(LOAD);
		return "error";
	    }
            chop;
	    if ( ! m/\t/ ) {
		s/:/\t/g;
		$_ = &fix_splits($_);
	    }
            &create_trans($_);
	}
    }

    close(LOAD);

    &calc_trans();

    return "ok";
}

# import cbb 0.8 version file
# there must already be an account open.  don't init the 
# current account
sub import_cbb9 {
    # in: filename
    # out: result

    my($file) = @_;
    my($file_version) = "";
    my($junk);

    $CBB::sorted_keys = 0;
    $CBB::calced = 0;

    print DEBUG "Loading the cbb format file: $file\n" if $CBB::debug;

    if ( $CBB::decrypt ne "" ) {
	open(LOAD, "$CBB::decrypt < $file|") || return "error";
    } else {
    	open(LOAD, "<$file") || return "error";
    }

    while ( <LOAD> ) {
	if ( m/^#/ ) {
	    # toss the comment (but first check for any goodies.)
	    if ( m/version/i ) {
		($junk, $junk, $junk, $file_version) = split;
		print DEBUG "Data file version = $file_version\n" if $CBB::debug;
	    }
	} else {
	    if ( $file_version eq "") {
		print DEBUG "no data file version, file encrypted ?" if $CBB::debug;
		close(LOAD);
		return "error";
	    }
            chop;
	    if ( ! m/\t/ ) {
		s/:/\t/g;
		$_ = &fix_splits($_);
	    }
            &create_trans($_);
	}
    }

    close(LOAD);

    &calc_trans();

    return "ok";
}

sub OSsource {

    my ($file) = $_[0];
    my ($Nn, $Nr, $buffer, $char, $type);

    open (FILE, "$file");
    read FILE, $buffer, 1024;
    close (FILE);
    
    $Nn = 0;
    $Nr = 0;
    while (length ($buffer) > 0) {
	$char = chop ($buffer);
	if ($char eq "\n") {
	    $Nn ++;
	}	
	if ($char eq "\r") {
	    $Nr ++;
	}	
    }
    
    if ($Nn && $Nr) {
	$type = "DOS";
    }
    
    if ($Nn && !$Nr) {
	$type = "UNIX";
    }
    
    if (!$Nn && $Nr) {
	$type = "MAC";
    }
    
    if (!$Nn && !$Nr) {
	$type = "UNKNOWN";
    }

    return $type;
}

sub fix_splits {
    # in: transaction with old two field per record splits
    # out: transaction with new three field per record splits

    my($line) = @_;
    my($date, $check, $desc, $debit, $credit, $cat, $com, $cleared, $total) =
	split(/\t/, $line);
    my(@cats, $i, $max, $newcat);

    if ( $cat =~ m/\|/ ) {
        @cats = split(/\|/, $cat);

        $i = 0;
        $max = ($#cats - 1) / 2;
        $newcat = "|";

        while ( $i < $max ) {
    	    $newcat .= $cats[$i * 2 + 1] . "||" . 
		       $cats[$i * 2 + 2] . "|";
	    $i++;
        }
    } else {
	$newcat = $cat;
    }

    return "$date\t$check\t$desc\t$debit\t$credit\t$newcat\t$com\t$cleared\t$total";
}

1;				# need to return a true value
